/*
    $Id: indices.h 80 2010-02-21 16:20:54Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _INDICES_H
#define _INDICES_H

#include <stdexcept>
#include <list>
#include <vector>
#include <iterator>

#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/tuple/tuple.hpp>

#include "typesdef.h"
#include "io.h"

template<class Iterator,int N> class tuple_iterator:
	public std::iterator<
		typename std::iterator_traits<Iterator>::iterator_category,
		typename boost::tuples::element<N, typename std::iterator_traits<Iterator>::value_type >::type,
		typename std::iterator_traits<Iterator>::difference_type > {
public:
	typedef Iterator subiterator_type;
	
	tuple_iterator( subiterator_type it ): it_(it) {
	}
	tuple_iterator( const tuple_iterator& it ): it_(it.it_) {
	}
	tuple_iterator& operator++() {
		++it_;
		return *this;
	}
	tuple_iterator& operator++(int) {
		it_++;
		return *this;
	}

	bool operator==(const tuple_iterator& rhs) {
		return it_==rhs.it_;
	}
	bool operator!=(const tuple_iterator& rhs) {
		return it_!=rhs.it_;
	}
	typename std::iterator_traits<Iterator>::difference_type operator-(const tuple_iterator& rhs) {
		return it_ - rhs.it_;
	}
	typename boost::tuples::access_traits<
		typename boost::tuples::element<N, typename std::iterator_traits<Iterator>::value_type >::type >::const_type operator*() const {
		return boost::tuples::template get<N>( *it_ );
	}
private:
	subiterator_type it_;
};

namespace indices {
	using namespace boost::numeric::ublas;

	typedef double                value_type;
	typedef vector< value_type >  vector_type;
	typedef vector_type::size_type size_type;
	typedef symmetric_matrix< value_type > symmetric_matrix_type;
	typedef vector_type           flux_type;
	typedef symmetric_matrix_type vars_type;
	typedef symmetric_matrix_type rho1_type;
	typedef vector_type           rho2_type;
	typedef vector_type           indesis_type;
	typedef symmetric_matrix_type indices_type;

	flux_type flux_nonlinearity_correction( const flux_type& flux, const vector_type& deadtime );
	flux_type flux_background_correction( const flux_type& flux, const vector_type& scatter, const vector_type& background );
	boost::tuple< vars_type, rho1_type, rho2_type >
		vars_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2, const vector_type& deadtime, const vector_type& npoisson );
	boost::tuple< indices_type, indesis_type > calculate( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 );
	
};

class indices_factory;
class indices_abstract_filter;

class indices_factory {
	public:
		typedef indices::value_type  value_type;
		typedef indices::vector_type vector_type;
		typedef indices::size_type   size_type;
		typedef indices::flux_type   flux_type;
		typedef indices::indices_type indices_type;
		typedef indices::indesis_type indesis_type;

		typedef indices_abstract_filter filter_type;
		typedef std::list< filter_type* >  filter_list_type;

		typedef boost::posix_time::ptime time_type;
		
		typedef boost::tuple< time_type, flux_type, indices_type, indesis_type > sample_type;
		typedef std::vector< sample_type > samples_list_type;

		typedef samples_list_type::const_iterator const_iterator;
		typedef tuple_iterator<const_iterator, 1> flux_const_iterator;
		typedef tuple_iterator<const_iterator, 2> indices_const_iterator;
		typedef tuple_iterator<const_iterator, 3> indesis_const_iterator;
	private:
		filter_list_type  m_filter_list;
		samples_list_type m_samples_list;
	public:
		indices_factory();
		indices_factory( const indices_factory& factory );
		void swap(indices_factory& factory) throw();
		indices_factory& operator=( const indices_factory& factory );
		~indices_factory();

		template<class F> void add_filter( const F& filter ) {
			m_filter_list.push_back( new F(filter) );
		}

		bool flux_filter( const time_type& time, const flux_type& flux, bool wexcept = true ) const;
		
		void accumulate( const time_type& time, const flux_type& flux, const indices_type& indices, const indesis_type& indesis );

		size_type      size()  const { return m_samples_list.size(); }
		const_iterator begin() const { return m_samples_list.begin(); }
		const_iterator end()   const { return m_samples_list.end(); }
		flux_const_iterator flux_begin() const { return m_samples_list.begin(); }
		flux_const_iterator flux_end()   const { return m_samples_list.end(); }
		indices_const_iterator indices_begin() const { return m_samples_list.begin(); }
		indices_const_iterator indices_end()   const { return m_samples_list.end(); }
		indesis_const_iterator indesis_begin() const { return m_samples_list.begin(); }
		indesis_const_iterator indesis_end()   const { return m_samples_list.end(); }
};

class indices_abstract_filter {
	public:
		typedef indices_factory::value_type  value_type;
		typedef indices_factory::vector_type vector_type;
		typedef indices_factory::size_type   size_type;

		typedef indices_factory::flux_type    flux_type;
		typedef indices_factory::indices_type indices_type;
		typedef indices_factory::indesis_type indesis_type;

		typedef indices_factory::time_type    time_type;
	private:
	public:
		indices_abstract_filter();
		virtual ~indices_abstract_filter() = 0;

		virtual bool flux_filter( const time_type& time, const flux_type& flux, bool wexcept = true ) const = 0;
		virtual indices_abstract_filter* clone() const = 0;
};

class indices_threshold_filter: public indices_abstract_filter {
	public:
		typedef indices_abstract_filter::value_type  value_type;
		typedef indices_abstract_filter::vector_type vector_type;
		typedef indices_abstract_filter::size_type   size_type;

		typedef indices_factory::flux_type    flux_type;
		typedef indices_factory::indices_type indices_type;
		typedef indices_factory::indesis_type indesis_type;
	private:
		value_type m_threshold;
	public:
		indices_threshold_filter( value_type threshold );
		~indices_threshold_filter();

		bool flux_filter( const time_type& time, const flux_type& flux, bool wexcept = true ) const;
		indices_abstract_filter* clone() const;
};

#endif // _INDICES_H
