/*
    $Id: profile.h 80 2010-02-21 16:20:54Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _PROFILE_H
#define _PROFILE_H

#include <string>
#include <stdexcept>

#include <lsp/cholesky_decomposition.h>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

#include "weif.h"

using namespace boost::numeric::ublas;

template<class VE> typename VE::size_type zero_elements( const vector_expression<VE>&x ) {
    typename VE::size_type n = 0;
    for( typename VE::size_type j = 0; j < x().size(); j++ )
	    n += ( x()(j) == 0 || std::isinf(x()(j)) || std::isnan(x()(j)) ) ? 1 : 0;
    return n;
}

template<class T> vector<T> element_invert( const vector<T>&x ) {
    vector<T> y = scalar_vector<T>(x.size(),1);
    for( unsigned int j = 0; j < x.size(); j++ ) y(j) /= x(j);
    return y;
}

// template<class T, class Tag> vector<T> diag_elements( const symmetric_matrix<T, Tag>&x ) {
//     typename matrix<T>::size_type r = x.size1();
//     slice s(0, 1, r );
//     return matrix_vector_slice< const symmetric_matrix<T, upper > > ( x, s, s );
// }
template<class ME> vector< typename ME::value_type > diag_elements( const matrix_expression<ME> &x ) {
    typename ME::size_type r = std::min( x().size1(), x().size2() );
    slice s(0, 1, r );
    return matrix_vector_slice< const ME > ( x, s, s );
}

class restoration_functor {
	public:
		struct profile_struct {
			vector<double> cn2_;
			vector<double> residual_;
			double chi2_;

			inline const double chi2() { return chi2_; }
			inline const vector<double>& cn2() { return cn2_; }
			inline const vector<double>& residual() { return residual_; }
		};
		typedef profile_struct result_type;

		typedef vector<double> grid_type;
		typedef matrix<double> weight_matrix;
	private:
		grid_type     m_grid;
		weight_matrix m_wmatrix;
		symmetric_matrix< double > m_cmatrix;
		matrix< double > m_pmatrix;
		double        m_secz;
		double        m_lambda_eff;

	public:
		static vector< double > flat_reorder( const matrix<double>& indices );
	
		restoration_functor( const vector<double>& grid, const symmetric_matrix< double >& corr_matrix, double secz, const Weights& basewf );

		template<class Iterator> static symmetric_matrix< double > calculate_covariation_matrix(Iterator begin, Iterator end) {
			typedef Iterator iterator_type;
			typedef typename iterator_type::difference_type size_type;
			
			if( begin == end ) throw std::out_of_range("Too few indices in weight matrix calculating.");

			vector< double > curr = flat_reorder( *begin );
			vector< double > inds = curr;
			symmetric_matrix< double, upper > corr = outer_prod( curr, curr );

			size_type num = 1;
			for( iterator_type it = ++begin; it != end; ++it,++num ){
				double k1 = double( num ) / double( num + 1 );
				double k2 =           1.0 / double( num + 1 );

				curr = flat_reorder( *it );

				inds = inds * k1 + curr * k2;
				corr = corr * k1 + outer_prod( curr, curr ) * k2;
			}

			corr -= outer_prod( inds, inds );

			const double damper = 0.8;
			
			triangular_adaptor< symmetric_matrix< double, upper >, unit_upper >( corr ) *= damper;

			if( zero_elements( diag_elements( corr ) ) > 0 ) {
				throw std::logic_error( (boost::format("Zero variances: %1% x %2% %3% %4%") % corr.size1() % corr.size2() % inds % vector<double>( diag_elements( corr )) ).str().c_str() );
			}

			return corr;
		}

		result_type operator() ( const matrix<double>& indices, double secz ) const;

		double seeing( const vector<double>& cn2 ) const;
		vector<double> indices_residual( const vector<double>& residuals ) const;
};

#endif // _PROFILE_H
