/*
    $Id: typesdef.h 80 2010-02-21 16:20:54Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _TYPESDEF_H
#define _TYPESDEF_H
#include <vector>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include <yalpa/spheric.h>

#include "lambert.h"

struct object {
	unsigned int hd;
	double       vmag;
	double       bvind;
	std::string  name;
	yalpa::equatorial<double> coordinates;
	std::string  sclass;
};

class datum {
public:
    double  mean;
    double  errs;
    datum( double m = 0.0, double e = 0.0 ) : mean(m), errs(e) {}
};

using namespace boost::numeric::ublas;

template<class OP, class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, OP>::result_type apply_to_all (const vector_expression<E> &e, const OP& op = OP() ) {
    typedef typename vector_unary_traits<E, OP>::expression_type expression_type;
    return expression_type (e ());
}
template<class OP, class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, OP>::result_type apply_to_all (const matrix_expression<E> &e, const OP& op = OP() ) {
    typedef typename matrix_unary1_traits<E, OP>::expression_type expression_type;
    return expression_type (e ());
}

template<class T> class nocase_less {
	public:
		typedef T string_type;
	private:
		static bool nocase_cmp( typename string_type::value_type c1, typename string_type::value_type c2 ) {
			return std::toupper( c1 ) < std::toupper( c2 );
		}
	public:
		nocase_less() {}
		/* it is a less function */
		bool operator()( const string_type& s1, const string_type& s2 ) const {
			return std::lexicographical_compare( s1.begin(), s1.end(), s2.begin(), s2.end(), &nocase_cmp );
		}
};

namespace functor {
	
    template <class T> class inv {
        public:
        typedef T value_type;
        typedef T result_type;
        inv() { }
        static  result_type apply( const value_type& x) {
            return ( x != value_type(0) )? value_type(1)/x : value_type(0);
        }
    };
    template <class T> class exp {
        public:
        typedef T value_type;
        typedef T result_type;
        exp() { }
        static  result_type apply( const value_type& x) {
            return std::exp( x );
        }
    };
    template <class T> class log {
        public:
        typedef T value_type;
        typedef T result_type;
        log() { }
        static  result_type apply( const value_type& x) {
            return std::log( x );
        }
    };
    template <class T> class lambertW {
        public:
        typedef T value_type;
        typedef T result_type;
        lambertW() { }
        static result_type apply( const value_type& x) {
            return lambert_W0(x);
        }
    };
    template <class T> class sqrt {
        public:
        typedef T value_type;
        typedef T result_type;
        sqrt() { }
        static  result_type apply( const value_type& x) {
            return std::sqrt(x);
        }
    };
    template <class T> class m1mx {
        public:
        typedef T value_type;
        typedef T result_type;
        m1mx() { }
        static result_type apply( const value_type& x) {
            return value_type(1) - x;
        }
    };
    template <class T> class m1px {
        public:
        typedef T value_type;
        typedef T result_type;
        m1px() { }
        static result_type apply( const value_type& x) {
            return value_type(1) + x;
        }
    };
}

template<class ME> matrix_vector_slice< const ME > diag_elements( const ME& x ){
    typename ME::size_type r = std::min( x.size1(), x.size2() );
    slice s(0, 1, r);
    return matrix_vector_slice< const ME >( x, s, s );
}

template< int p, class T > vector<T> mass_order( const vector<T>&x ) {
    vector<T> y( x.size()*(x.size()+1)/2);
    subrange( y, 0, x.size() ) = x;
    unsigned int k = x.size();
    for( unsigned int i = 0; i < x.size()-1; i++ )
        for( unsigned int j = i+1; j < x.size(); j++ ) {
            if( p == 1 ) y(k++) = x(i);
            if( p == 2 ) y(k++) = x(j);
        }
    return y;
}

template<class T> vector<T> mass_order_prod( const vector<T>& x ) {
    return  element_prod( mass_order<1>( x ), mass_order<2>( x ) );
}

template<class T> vector<T> mass_order_summ( const vector<T>& x ) {
    return  mass_order<1>( x ) + mass_order<2>( x );
}

template<class T> struct average_type_traits;

namespace {
	template<class Iterator,class T> std::pair<T,T> ublas_average( Iterator begin, Iterator end ) {
		typedef T        container_type;
		typedef Iterator iterator_type;
		typedef typename container_type::size_type  size_type;
		typedef typename container_type::value_type value_type;
		typedef std::pair< container_type, container_type > result_type;
		
		if( begin == end ) return result_type();

		result_type ret( *begin, element_prod( *begin, *begin ) );
		size_type num = 1;
		for( iterator_type it = ++begin; it != end; ++it,++num ) {
			double k1 = double( num ) / double( num + 1 );
			double k2 =           1.0 / double( num + 1 );

			ret.first  = ret.first  * k1 + *it * k2;
			ret.second = ret.second * k1 + element_prod( *it, *it ) * k2;
		}

		ret.second -= element_prod( ret.first, ret.first );
		if( num > 1 )
			ret.second = element_div( apply_to_all( ret.second / double(num-1), functor::sqrt<value_type>() ), ret.first );
		else
			ret.second = container_type();
		return ret;
	}

	template<class Iterator,class T> std::pair<T,T> pod_average( Iterator begin, Iterator end ) {
		typedef T value_type;
		typedef Iterator iterator_type;
		typedef std::pair< value_type, value_type > result_type;
		typedef typename iterator_type::difference_type size_type;

		if( begin == end ) return result_type();

		result_type ret( *begin, *begin * (*begin) );
		size_type num = 1;
		for( iterator_type it = ++begin; it != end; ++it,++num ) {
			double k1 = double( num ) / double( num + 1 );
			double k2 =           1.0 / double( num + 1 );

			ret.first  = ret.first  * k1 + *it * k2;
			ret.second = ret.second * k1 + *it * ( *it ) * k2;
		}

		ret.second -= ret.first * ret.first;
		if( num > 1 )
			ret.second = std::sqrt( ret.second / double(num-1) ) / ret.first;
		else
			ret.second = value_type(0);
		return ret;
	}
}

template<class U> struct average_type_traits< vector< U > > {
	typedef vector< U > value_type;
	typedef std::pair< value_type, value_type > result_type;
	
	template<class Iterator> static result_type average( Iterator begin, Iterator end ) {
		return ublas_average<Iterator, value_type>(begin,end);
	}
};

template<class U> struct average_type_traits< symmetric_matrix< U > > {
	typedef symmetric_matrix< U > value_type;
	typedef std::pair< value_type, value_type > result_type;

	template<class Iterator> static result_type average( Iterator begin, Iterator end ) {
		return ublas_average<Iterator, value_type>(begin,end);
	}
};

template<> struct average_type_traits<double> {
	typedef double value_type;
	typedef std::pair< value_type, value_type > result_type;

	template<class Iterator> static result_type average( Iterator begin, Iterator end ) {
		return pod_average<Iterator, value_type>(begin,end);
	}
};

template<class Iterator> typename average_type_traits< typename Iterator::value_type >::result_type average( Iterator begin, Iterator end ) {
	typedef average_type_traits< typename Iterator::value_type > traits_type;
	return traits_type::average( begin, end );
}

struct is_positive_number {
	bool operator()( const double& x ) {
		return ! ( std::isnan( x ) || std::isinf( x ) || ( x < 0 ) );
	}
};

#endif // _TYPESDEF_H
